from .resource import Resource
from .ref import Ref
from .event import Event
from .relation import Relation


class Key(Resource):

    """
    From the Orchestrate docs:

    "Key/Value is core to Orchestrate.io. 
    All other query types are built around this data type. 
    Key/Value pairs are pieces of data
    identified by a unique key for a collection
    and have corresponding value."

    The key object represents a single key/value in Orchestrate,
    and provides methods for working with refs, events, and relations,
    as well as creating, reading, updating, and destroying key/values.

    ```python
    key = collection.key(NAME)
    # create
    response = key.put({"hello": "world"}).result()
    response.raise_for_status()
    # create a ref to work with this version of the key/value
    ref_value = response.path['ref']
    ref = key.ref(ref_value)
    # read
    response = key.get().result()
    print response.json()
    # {"hello": "world"}
    # update
    response = key.put({"goodnight": "moon"}, **{"If-Match": ref_value}).result()
    response.raise_for_status()
    ref_value = response.path['ref']
    # delete
    response = key.delete(**{"If-Match": ref_value}).result()
    response.raise_for_status()
    ```
    """

    def ref(self, name, **kwargs):
        """
        Returns an object for dealing with a given Ref value.
        See `Ref` for more info.
        """
        return self._init_child(Ref, 'refs', name, **kwargs)

    def event(self, event_type, **kwargs):
        """
        Returns an object for creating and querying events.
        See `Event` for more info.
        """
        return self._init_child(Event, 'events', event_type, **kwargs)

    def relation(self, **kwargs):
        """
        Returns an object for creating and querying relations.
        See `Relation` for more info.
        """
        return self._init_child(Relation, **kwargs)

    def graph(self, **kwargs):
        """
        Alias to `relation`
        """
        return self.relation(**kwargs)



from .resource import Resource
from collections import Iterator
from requests_futures.sessions import FuturesSession


class Page(Iterator, Resource):

    """
    Class used for paging through search results.

    Can be used as an iterator, ex: `for page in Page`,
    or by using the methods `next` and `prev` explicitly
    to page through results.

    Used in searching collections and events.
    """

    def __init__(self, uri, querydict={}, **kwargs):
        opts = dict(kwargs, params=querydict)
        # if async, create a callback that stores current page state after each request
        if 'session' in kwargs and type(kwargs['session']) == FuturesSession:
            opts['background_callback'] = self._handle_res
        super(Page, self).__init__(uri, **opts)
        self._url_root = self.uri[:self.uri.find('/v0')]
        self.response = None

    def _handle_res(self, session, response):
        """
        Stores the response, which we use for determining
        next and prev pages.
        """
        self.response = response

    def reset(self):
        """
        Clear the page's current place.

            page_1 = page.next().result()
            page_2 = page.next().result()
            page.reset()
            page_x = page.next().result()
            assert page_x.url == page_1.url
        """
        self.response = None

    def _handle_page(self, querydict={}, val='next', **headers):
        """
        Executes the request getting the next (or previous) page,
        incrementing (or decrementing) the current page.
        """
        # if async, wait for previous page to load
        if hasattr(self.response, 'result'):
            self.response.result()
        # update uri based on next page
        if self.response:
            self.response.raise_for_status()
            _next = self.response.json().get(val, False)
            if _next:
                self.uri = self._url_root + _next
            else:
                raise StopIteration
        # execute request
        response = self.get(querydict, **headers)
        self._handle_res(None, response)
        return response

    def next(self, querydict={}, **headers):
        """
        Gets the next page of results.
        Raises `StopIteration` when there are no more results.
        """
        return self._handle_page(querydict, **headers)

    def __next__(self):
        return self.next()

    def prev(self, querydict={}, **headers):
        """
        Gets the previous page of results.
        Raises `StopIteration` when there are no more results.

        Note: Only collection searches provide a `prev` value.
        For all others, `prev` will always return `StopIteration`.
        """
        return self._handle_page(querydict, 'prev', **headers)        



from . import resource
from . import collection


class Client(resource.Resource):

    """
    The top-level resource for interacting with Orchestrate.
    Handles authentication, spawns collections.

    ```
    client = porc.Client(API_KEY)
    client.ping().result().raise_for_status()
    # if it throws an error, you're not authenticated
    ```
    """

    def __init__(self, api_key, api_version=0, **kwargs):
        uri = 'https://%s:@api.orchestrate.io/v%d' % (api_key, api_version)
        super(Client, self).__init__(uri, **kwargs)

    def collection(self, name, **kwargs):
        """
        Spawns a collection object, inheriting options from the client object.
        Note that this does not create a collection in Orchestrate.
        """
        return self._init_child(collection.Collection, name, **kwargs)

    def ping(self):
        """
        Sends Orchestrate a HEAD request to determine
        connectivity and authentication.
        """
        return self.head()



from .resource import Resource
from .client import Client
from .collection import Collection
from .page import Page
from .key import Key
from .ref import Ref
from .event import Event
from .relation import Relation
from . import util


from .resource import Resource
from .page import Page


class Relation(Resource):

    """
    From the Orchestrate docs:

    "The Graph functionality allows for directed relations 
    to be created between collection/key pairs 
    and for those relations to be traversed."

    A relation object allows you to create, delete, and query
    relations associated with a key.

    ```python
    relation = key.relation()
    # create
    response = relation.put('loves', 'people', 'David Bowie').result()
    response.raise_for_status()
    # query
    response = relation.get('loves')
    response.raise_for_status()
    print response.json()
    # {
    #     "count": 1,
    #     "results": [
    #         {
    #             "path": {
    #                 "collection": "people",
    #                 "key": "David Bowie",
    #                 "ref": "0acfe7843316529f"
    #             },
    #             "value": {
    #                 "age": 67,
    #                 "name": "David Bowie"
    #             }
    #         }
    #     ]
    # }
    # delete
    response = relation.delete('loves', 'people', 'David Bowie').result()
    response.raise_for_status()
    ```
    """

    def __init__(self, uri, **kwargs):
        self._root_url = uri
        super(Relation, self).__init__(uri, **kwargs)

    def get(self, *relations, **headers):
        """
        Queries all relations specified in `relations`, ex:

            relation.get('friend', 'friend')

        ...would return all friends of friends.

        N.B.: Graph query results are not currently paginated.
        """
        resource = self._init_child(Resource, 'relations', *relations)
        return resource.get(**headers)

    def put(self, relation, collection, key, **headers):
        """
        Creates a relationship between two objects.
        Relations can span collections.
        """
        resource = self._init_child(Resource, 'relation', relation, collection, key)
        return resource.put(**headers)

    def delete(self, relation, collection, key, querydict={}, **headers):
        """
        Deletes a relationship between two objects.
        Automatically sets `?purge=true` in the querystring.
        """
        resource = self._init_child(Resource, 'relation', relation, collection, key)
        querydict['purge'] = True
        return resource.delete(querydict, **headers)



from .resource import Resource


class Ref(Resource):

    """
    From the Orchestrate docs:

    "Refs are used to identify specific immutable values that have been assigned to keys."

    Ref objects allow you to work with specific versions of a doc,
    particularly past versions.

    ```python
    ref = key.ref(ref_value)
    response = ref.get().result()
    print response.json()
    # {"hello": "world"}
    ```
    """
    # everything is defined by the parent :D
    pass



import json
import requests
from requests_futures.sessions import FuturesSession
import copy
import re
try:
    # python 2
    from urllib import quote
except ImportError:
    # python 3
    from urllib.parse import quote

URL_PATTERNS = [
    "(?P<collection>.+)/(?P<key>.+)/relations/(?P<kinds>.+)",
    "(?P<collection>.+)/(?P<key>.+)/events/(?P<type>.+)",
    "(?P<collection>.+)/(?P<key>.+)/refs/(?P<ref>.+)",
    "(?P<collection>.+)/(?P<key>.+)",
    "(?P<collection>.+)"
]

RESPONSE_PATTERNS = [
    "/v0/(?P<collection>.+)/(?P<key>.+)/events/(?P<type>.+)/(?P<timestamp>\d+)/(?P<ordinal>\d+)",
    "/v0/(?P<collection>.+)/(?P<key>.+)/refs/(?P<ref>.+)"
]

class Resource(object):

    """
    Ancestor for all Orchestrate resources.
    Handles storing URL state, headers, sessions, etc.

    If you create an object, like a `Client`, then use that to
    create a `Collection` object, the `Collection` will inherit any options
    from the `Client` object, such as auth settings.

    Likewise, all objects inheriting from Resource can make HTTP requests
    using the methods `head`, `get`, `put`, `post`, and `delete`,
    even if Orchestrate would only return a `405 Method Not Allowed`
    for a given method.
    """
    def __init__(self, uri, **kwargs):
        self.uri = uri

        # if given an existing session, use it
        if kwargs.get('session'):
            self._session = kwargs['session']
            del kwargs['session']
        # set the resource's session as either sync or async
        # according to keyword arguments
        elif kwargs.get('async') == False:
            self._session = requests.Session()
            del kwargs['async']
        # default to async
        else:
            self._session = FuturesSession()

        # set path parameters according to given url
        self._set_path()

        # pass remaining kwargs as defaults to requests
        self._set_options(**kwargs)

    def _init_child(self, child_obj, *path, **kwargs):
        """
        Create child objects from the resource's current settings.
        Used to create descendent resources, ex: keys from collections
        """
        if len(path):
            uri = self.uri + '/' + '/'.join([quote(path_elem) for path_elem in path])
        else:
            uri = self.uri

        # merge current and new settings into a new dict
        opts = dict(self.opts, **kwargs)
        # if both the parent and child have a callback, wrap them together
        if 'background_callback' in self.opts and 'background_callback' in kwargs:
            def callback(session, response):
                self.opts['background_callback'](session, response)
                kwargs['background_callback'](session, response)
            opts['background_callback'] = callback

        return child_obj(uri, session=self._session, **opts)

    def _set_path(self):
        """
        Set the path variable with params like $collection, etc.
        based on the object's URI
        """
        start_after_segment = 'api.orchestrate.io/v0/'
        start_after_index = self.uri.find(
            start_after_segment) + len(start_after_segment)
        path = self.uri[start_after_index:]
        for regex in URL_PATTERNS:
            match = re.match(regex, path)
            if match:
                self.path = match.groupdict()
                # handle relations with multiple kind parameters
                if 'kinds' in self.path:
                    self.path['kinds'] = self.path['kinds'].split('/')
                break
        else:
            self.path = dict()

    def _set_options(self, **kwargs):
        """
        Sets header defaults and updates them with given kwargs
        """
        # set defaults
        if not hasattr(self, 'opts'):
            self.opts = {
                'headers': {
                    'Content-Type': 'application/json',
                    'Accept': 'application/json'
                }
            }
        # merge passed kwargs into opts
        if kwargs:
            self.opts.update(kwargs)

    def _make_request(self, method, body={}, **headers):
        """
        Executes the request based on the given body and headers
        along with options set on the object.
        """
        # create a copy of options so we can mess with it
        opts = copy.copy(self.opts)

        # normalize `params` kwarg according to method
        if body:
            if method in ['head', 'get', 'delete']:
                if type(body) == dict:
                    # convert True and False to true and false
                    for key, value in list(body.items()):
                        if value == True:
                            body[key] = 'true'
                        elif value == False:
                            body[key] = 'false'
                opts['params'] = body
            else:
                opts['data'] = json.dumps(body)

        # auto-quote headers like if-match
        for header in ['If-Match', 'If-None-Match']:
            if header in headers:
                # wrap in quotes if not already surrounded
                if headers[header][0] != '"':
                    headers[header] = '"%s"' % headers[header]
        # update opts with headers following any modifications
        opts['headers'].update(headers)

        # make the request
        if type(self._session) == FuturesSession:
            # handle multiple background_callbacks
            if 'background_callback' in opts:
                background_callback = opts['background_callback']
                def callback(session, response):
                    background_callback(session, response)
                    self._handle_response(session, response)
                del opts['background_callback']
            else:
                callback = self._handle_response
            future = getattr(self._session, method)(
                self.uri, background_callback=callback, **opts)
            return future
        else:
            response = self._session.request(method, self.uri, **opts)
            self._handle_response(None, response)
            return response

    def _handle_response(self, session, response):
        """
        Modifies the response, normalizing headers and the like
        """
        response.path = dict()
        for regex in RESPONSE_PATTERNS:
            # check location
            location_match = re.match(
                regex, response.headers.get('location', ''))
            # if not in location, try content-location
            if not location_match:
                location_match = re.match(
                    regex, response.headers.get('content-location', ''))
            # else, try the etag
            if not location_match:
                location_match = re.match('"(?P<ref>.+)-gzip"', response.headers.get('ETag', ''))
            # if a match was found in either place, attach it to the `path`
            # attribute
            if location_match:
                response.path = location_match.groupdict()
                break

    def head(self, querydict={}, **headers):
        """
        Make a HEAD request against the object's URI.
        `querydict` is a mapping object, like a dict, 
        that is serialized into the querystring.
        Any keyword arguments are passed along as headers.
        """
        return self._make_request('head', querydict, **headers)

    def get(self, querydict={}, **headers):
        """
        Make a GET request against the object's URI.
        `querydict` is a mapping object, like a dict, 
        that is serialized into the querystring.
        Any keyword arguments are passed along as headers.
        """
        return self._make_request('get', querydict, **headers)

    def put(self, body={}, **headers):
        """
        Make a PUT request against the object's URI.
        `querydict` is a mapping object, like a dict, 
        that is serialized into the querystring.
        Any keyword arguments are passed along as headers.
        """
        return self._make_request('put', body, **headers)

    def post(self, body={}, **headers):
        """
        Make a POST request against the object's URI.
        `querydict` is a mapping object, like a dict, 
        that is serialized into the querystring.
        Any keyword arguments are passed along as headers.
        """
        return self._make_request('post', body, **headers)

    def delete(self, querydict={}, **headers):
        """
        Make a DELETE request against the object's URI.
        `querydict` is a mapping object, like a dict, 
        that is serialized into the querystring.
        Any keyword arguments are passed along as headers.
        """
        return self._make_request('delete', querydict, **headers)



from . import resource
from . import key
from . import page


class Collection(resource.Resource):

    """
    From the Orchestrate docs:
    
    "Collections are groupings of the JSON objects. 
    Collections are analogous to tables in a relational database."

    The collection object performs searches over the JSON objects stored in a collection, 
    and can spawn key objects representing JSON objects.

    ```python
    collection = porc.Client(API_KEY).collection(NAME)
    for page in collection.search({"query": "cool_*"}):
        print page.json()
        # {
        #     "count": 1,
        #     "next": "/v0/collection?limit=10&query=test&offset=1",
        #     "prev": "/v0/collection?limit=10&query=test&offset=0",
        #     "results": [
        #         {
        #             "path": {
        #                 "collection": "neat_stuff",
        #                 "key": "cool_beans",
        #                 "ref": "20c14e8965d6cbb0"
        #             },
        #             "score": 1.0,
        #             "value": {
        #                 "coolness": 999
        #             }
        #         }
        #     ],
        #     "total_count": 100
        # }
    ```
    """

    def key(self, name, **kwargs):
        """
        Spawns a key object, inheriting options from the collection object.
        Note that this does not create a key in Orchestrate.
        """
        return self._init_child(key.Key, name, **kwargs)

    def list(self, **kwargs):
        """
        Returns a page object to page through list results.
        """
        return page.Page(self.uri, kwargs, session=self._session, **self.opts)

    def search(self, **kwargs):
        """
        Returns a page object to page through search results.

        Alias to `self.list`.
        """
        return self.list(**kwargs)

    def delete(self, querydict={}, **headers):
        """
        Deletes a collection and all the keys it contains.
        Automatically sets `?force=true` in the querystring.
        """
        querydict['force'] = True
        return super(Collection, self).delete(querydict, **headers)



from .resource import Resource
from .page import Page
from .util import create_timestamp
from datetime import datetime

class Event(Resource):

    """
    From the Orchestrate docs:

    "Events are a way to associate time-ordered data with a key."

    The event object allows you to create, read, update, and delete
    different events, and to perform queries over events associated
    with a key.

    ```python
    event = key.event(TYPE)
    # create
    response = event.post({"msg": "hello world, again"}).result()
    timestamp = response.path['timestamp']
    ordinal = response.path['ordinal']
    response.raise_for_status()
    # read
    response = event.get(timestamp, ordinal).result()
    response.raise_for_status()
    print response.json()
    # {
    #     "path": {
    #         "collection": "asdf",
    #         "key": "key",
    #         "ref": "ae3dfa4325abe21e",
    #         "type": "played",
    #         "timestamp": 1369832019085,
    #         "ordinal": 9
    #     },
    #     "value": {
    #       "msg": "hello world, again"
    #     },
    #     "timestamp": 1369832019085,
    #     "ordinal": 9
    # }
    # update
    response = event.put(timestamp, ordinal, {"msg": "goodnight moon"}).result()
    response.raise_for_status()
    # query
    for page in event.list():
        print page.json()
        # {
        #   "results": [
        #     {
        #       "path": {
        #         "collection": "collection",
        #         "key": "key",
        #         "type": "type",
        #         "timestamp": 1369832019085,
        #         "ordinal": 9,
        #         "ref": "ae3dfa4325abe21e"
        #       },
        #       "value": {
        #         "msg": "goodnight moon"
        #       },
        #       "timestamp": 1369832019085,
        #       "ordinal": 9
        #     }
        #   "count": 1
        # }
    # delete
    response = event.delete(timestamp, ordinal).result()
    response.raise_for_status()
    ```
    """

    def get(self, timestamp=None, ordinal=None, querydict={}, **headers):
        """
        Retrieves a given event. 
        If timestamp is a datetime object, 
        it will be converted to milliseconds since epoch.
        """
        if not timestamp and not ordinal:
            return self.list(querydict, **headers)
        else:
            if type(timestamp) == datetime:
                timestamp = create_timestamp(timestamp)
            resource = self._init_child(Resource, str(timestamp), ordinal)
            return resource.get(**headers)

    def post(self, body, timestamp=None, **headers):
        """
        Creates a new event.
        If timestamp is not provided, Orchestrate will give it one.
        """
        if timestamp:
            if type(timestamp) == datetime:
                timestamp = create_timestamp(timestamp)
            resource = self._init_child(Resource, str(timestamp))
        else:
            resource = self._init_child(Resource)

        return resource.post(body, **headers)

    def put(self, timestamp, ordinal, body, **headers):
        """
        Updates a given event.
        If timestamp is a datetime object, 
        it will be converted to milliseconds since epoch.
        """
        if type(timestamp) == datetime:
            timestamp = create_timestamp(timestamp)
        resource = self._init_child(Resource, str(timestamp), ordinal)
        return resource.put(body, **headers)

    def delete(self, timestamp, ordinal, querydict={}, **headers):
        """
        Deletes a given event.
        Automatically sets `?purge=true` in the querystring.
        """
        if type(timestamp) == datetime:
            timestamp = create_timestamp(timestamp)
        resource = self._init_child(Resource, str(timestamp), ordinal)
        querydict['purge'] = True
        return resource.delete(querydict, **headers)

    def list(self, querydict, **headers):
        """
        Returns a page representing a given query of events.
        See `Page` for more details on the page object.
        """
        return Page(self.uri, querydict, session=self._session, headers=headers)



#!/usr/bin/env python

from setuptools import setup

setup(name='porc',
      version='0.2.0',
      description='Asynchronous Orchestrate.io Interface',
      author='Max Thayer',
      author_email='max@orchestrate.io',
      url='https://github.com/orchestrate-io/porc',
      packages=['porc'],
      license='MIT',
      install_requires=[
          'requests-futures==0.9.4',
          'vcrpy==1.0.2'
      ],
      test_suite="tests",
      classifiers=[
          'Intended Audience :: Developers',
          'Natural Language :: English',
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python',
          'Programming Language :: Python :: 2',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.3',
          'Programming Language :: Python :: 3.4'
      ],
      )
